/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/limits.hh"
#include "src/database/impl/is_keyset_ok_for_update.hh"

#include <libfred/registrable_object/keyset/check_dns_key.hh>

#include <liblog/liblog.hh>

#include <cctype>

namespace Fred {
namespace Akm {

namespace {

bool is_base64_data_character(char c)
{
    return ( c == '+' ) || ( c == '/' ) ||
            (( '0' <= c ) && ( c <= '9' )) ||
            (( 'A' <= c ) && ( c <= 'Z' )) ||
            (( 'a' <= c ) && ( c <= 'z' ));
}

bool is_base64_pad_character(char c)
{
    return c == '=';
}

bool is_base64_insignificant_allowed_character(char c)
{
    // see https://en.cppreference.com/w/cpp/string/byte/isspace#Notes
    return std::isspace(static_cast<unsigned char>(c));
}

bool is_correct(const DnskeyPublicKey& _public_key)
{
    unsigned len = 0;
    unsigned pads = 0;
    for (const char key_char : *_public_key)
    {
        if (is_base64_insignificant_allowed_character(key_char))
        {
            continue;
        }
        const bool no_pad_character_yet = (pads == 0);
        if (no_pad_character_yet && is_base64_data_character(key_char))
        {
            ++len;
            continue;
        }
        if (is_base64_pad_character(key_char))
        {
            static const unsigned max_number_of_pads = 2;
            if (max_number_of_pads <= pads)
            {
                return false;
            }
            ++len;
            ++pads;
            continue;
        }
        // neither data character nor pad nor space
        return false;
    }
    return (len % 4) == 0;
}

// DNS key item flags has only 3 allowed values: 0, 256, 257
bool is_correct(const DnskeyFlags& _flags)
{
    switch (*_flags)
    {
        case 0u:
        case 256u:
        case 257u:
            return true;
    }
    return false;
}

// DNS key item protocol has only 1 allowed value: 3
bool is_correct(const DnskeyProto& _proto)
{
    return *_proto == 3u;
}

bool is_correct(const LibFred::OperationContext& _ctx, const DnskeyAlg& _alg)
{
    const auto algorithm_usability = LibFred::DnsSec::get_algorithm_usability(_ctx, *_alg);
    return algorithm_usability == LibFred::DnsSec::Algorithm::usable;
}

bool has_invalid_dnskeys(const LibFred::OperationContext& _ctx, const Keyset& _keyset)
{
    for (const auto& dnskey : _keyset)
    {
        if (!is_correct(dnskey.flags) ||
            !is_correct(dnskey.proto) ||
            !is_correct(_ctx, dnskey.alg) ||
            !is_correct(dnskey.public_key))
        {
            return true;
        }
    }
    return false;
}

} // namespace Fred::Akm::Database::{anonymous}

} // namespace Fred::Akm
} // namespace Fred

bool Fred::Akm::Database::Impl::is_keyset_ok_for_update(const LibFred::OperationContext& _ctx, const Keyset& _keyset)
{
    if (_keyset.size() < Limit::keyset_size_min)
    {
        LIBLOG_DEBUG("attempting to add less than needed dnskeys");
        return false;
    }

    if (_keyset.size() > Limit::keyset_size_max)
    {
        LIBLOG_DEBUG("attempting to add more than needed dnskeys");
        return false;
    }

    if (has_invalid_dnskeys(_ctx, _keyset))
    {
        LIBLOG_DEBUG("new keyset has invalid dnskeys");
        return false;
    }
    return true;
}
