/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/database/impl/turn_off_dnssec.hh"
#include "src/exception.hh"

#include <libfred/poll/create_update_object_poll_message.hh>
#include <libfred/registrable_object/domain/update_domain.hh>

#include <liblog/liblog.hh>

namespace Fred {
namespace Akm {
namespace Database {
namespace Impl {

namespace {

bool is_special_delete_key(const Dnskey& dnskey)
{
    // RFC 8078 Section 4. "DNSSEC Delete Algorithm"
    static const std::string base64_encoded_zero = "AA==";
    return ((*dnskey.flags == 0) &&
            (*dnskey.proto == 3) &&
            (*dnskey.alg == 0) &&
            (*dnskey.public_key == base64_encoded_zero));
}

std::string make_exception_message(const LibFred::UpdateDomain::Exception& e)
{
    if (e.is_set_unknown_domain_fqdn())
    {
        return "unknown domain fqdn";
    }
    return "unknown error";
}

using LogEntryAkmTurnOff = Logger::LogEntry<Logger::LogEntryType::akm_turn_off>;

} // namespace Fred::Akm::Database::{anonymous}

bool is_dnssec_turn_off_requested(const Keyset& keyset)
{
    if (keyset.size() == 1)
    {
        const auto& the_only_key = *keyset.begin();
        if (is_special_delete_key(the_only_key))
        {
            return true;
        }
    }
    return false;
}

void turn_off_dnssec(
        const LibFred::OperationContext& _ctx,
        const LibFred::InfoDomainData& _info_domain_data,
        const KeysetMetadata& _keyset_metadata,
        std::function<std::unique_ptr<LogEntryAkmTurnOff>()> _make_log_entry)
{
    try
    {
        std::unique_ptr<LogEntryAkmTurnOff> log_entry;
        const auto domain_new_history_id = [&]()
        {
            LibFred::UpdateDomain update{_info_domain_data.fqdn, _keyset_metadata.registrar};
            update.unset_keyset();
            log_entry = _make_log_entry();
            if (log_entry != nullptr)
            {
                update.set_logd_request_id(*log_entry->get_id());
            }
            return update.exec(_ctx);
        }();
        LIBLOG_DEBUG("domain_new_history_id: {}", domain_new_history_id);
        LibFred::Poll::CreateUpdateObjectPollMessage{}.exec(_ctx, domain_new_history_id);
        LIBLOG_INFO("dnssec turned off");

        if (log_entry != nullptr)
        {
            const auto keyset_reference = KeysetReference{
                    KeysetId{_info_domain_data.keyset.get_value().id},
                    KeysetHandle{_info_domain_data.keyset.get_value().handle}};
            log_entry->close_on_success(keyset_reference);
        }
    }
    catch (const LibFred::UpdateDomain::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw ConfigurationError{"unknown registrar handle"};
        }
        LIBLOG_ERROR("failed to update domain: {}", make_exception_message(e));
        throw TurnOffDnssecFailed{e.what()};
    }
    catch (const LibFred::OperationException& e)
    {
        LIBLOG_ERROR("failed to create \"update object poll\" message: {}", e.what());
        throw TurnOffDnssecFailed{e.what()};
    }
    catch (const LibFred::InternalError& e)
    {
        LIBLOG_ERROR("internal error at create \"update object poll\" message: {}", e.what());
        throw TurnOffDnssecFailed{e.what()};
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unspecified error occurred : {}", e.what());
        throw TurnOffDnssecFailed{e.what()};
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error occurred");
        throw;
    }
}

} // namespace Fred::Akm::Database::Impl
} // namespace Fred::Akm::Database
} // namespace Fred::Akm
} // namespace Fred
