/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/export_scan_tasks.hh"
#include "src/exception.hh"
#include "src/scan_tasks.hh"
#include "src/grpc_client/export_scan_tasks.hh"

#include <libfred/db_settings.hh>
#include <libfred/opcontext.hh>
#include <liblog/liblog.hh>

namespace Fred {
namespace Akm {

namespace {

void add_is_auto_keyset(Database::ParamQuery& _sql, const KeysetMetadata& _keyset_metadata)
{
    _sql(" (r.handle = UPPER(").param_text(_keyset_metadata.registrar)(") "
            "AND UPPER(oregk.name) LIKE UPPER(").param_text(_keyset_metadata.handle_prefix + "%")(")) ");
}

Database::ParamQuery construct_export_query(const KeysetMetadata& _keyset_metadata, ExportScanType _export_scan_type)
{
    Database::ParamQuery sql;
    // clang-format off
    sql("SELECT oreg.uuid AS domain_uuid, oreg.name AS domain_fqdn, ns.fqdn AS nameserver, "
            "CASE "
                "WHEN d.keyset IS NULL THEN 'insecure' "
                "WHEN"); add_is_auto_keyset(sql, _keyset_metadata);
                    sql("THEN 'secure-auto' "
                "ELSE 'secure' "
            "END AS scan_type "
          "FROM host ns "
          "JOIN domain d ON d.nsset = ns.nssetid "
          "JOIN object_registry oreg ON oreg.id = d.id "
          "JOIN zone z ON z.id = d.zone ");
    // clang-format on

    switch (_export_scan_type)
    {
        case ExportScanType::insecure:
            sql("LEFT JOIN (object ok "
                      "JOIN object_registry oregk ON oregk.id = ok.id "
                      "JOIN registrar r ON r.id = ok.clid) ON ok.id = d.keyset "
                "WHERE d.keyset IS NULL");
            break;

        case ExportScanType::secure:
            // clang-format off
            sql("JOIN object ok ON ok.id = d.keyset "
                "JOIN object_registry oregk ON oregk.id = ok.id "
                "JOIN registrar r ON r.id = ok.clid "
               "WHERE NOT");
            add_is_auto_keyset(sql, _keyset_metadata);
            sql("AND oregk.erdate IS NULL "
                 "AND oregk.type = get_object_type_id('keyset')");
            // clang-format on
            break;

        case ExportScanType::secure_auto:
            // clang-format off
            sql("JOIN object ok ON ok.id = d.keyset "
                "JOIN object_registry oregk ON oregk.id = ok.id "
                "JOIN registrar r ON r.id = ok.clid "
               "WHERE");
            add_is_auto_keyset(sql, _keyset_metadata);
            sql(" AND oregk.erdate IS NULL "
                 "AND oregk.type = get_object_type_id('keyset')");
            // clang-format on
            break;

        case ExportScanType::all:
            // clang-format off
            sql("LEFT JOIN (object ok "
                      "JOIN object_registry oregk ON oregk.id = ok.id "
                      "JOIN registrar r ON r.id = ok.clid) ON ok.id = d.keyset "
               "WHERE oregk.erdate IS NULL "
                 "AND (oregk.type = get_object_type_id('keyset') "
                     "OR oregk.type IS NULL)");
            // clang-format on
            break;
    }

    // managed zone check
    sql(" AND z.fqdn IN (");
    Util::HeadSeparator in_separator("", ", ");
    for (const auto& zone : _keyset_metadata.zones)
    {
         sql(in_separator.get()).param_text(zone);
    }
    sql(") ");

    // outzone check
    // clang-format off
    sql("AND NOT EXISTS("
            "SELECT "
              "FROM object_state os "
              "JOIN enum_object_states eos "
                   "ON eos.id = os.state_id AND eos.name = 'outzone' "
             "WHERE os.object_id = oreg.id AND os.valid_to IS NULL"
        ")");
    // clang-format on
    return sql;
}
 
ScanTasks get_scan_tasks_from_query_result(const Database::Result& dbres)
{
    ScanTasks scan_tasks;
    for (unsigned int idx = 0; idx < dbres.size(); ++idx)
    {
        const auto row = dbres[idx];
        const auto domain_fqdn = DomainFqdn{static_cast<std::string>(row["domain_fqdn"])};
        const auto domain_uuid = from_string<DomainUuid>(static_cast<std::string>(row["domain_uuid"]));
        const auto scan_type = from_string<ScanType>(static_cast<std::string>(row["scan_type"]));
        const auto nameserver = Nameserver{static_cast<std::string>(row["nameserver"])};
        auto scan_tasks_iter = scan_tasks.find(domain_fqdn);
        if (scan_tasks_iter != end(scan_tasks))
        {
            if ((scan_tasks_iter->second.scan_type != scan_type) ||
                (scan_tasks_iter->second.uuid != domain_uuid))
            {
                struct DifferentData : ExportError
                {
                    explicit DifferentData() : ExportError{{}} { }
                    const char* what() const noexcept override { return "domain differs in scan type or uuid"; }
                };
                throw DifferentData{};
            }
            scan_tasks_iter->second.nameservers.insert(nameserver);
        }
        else
        {
            scan_tasks.insert(std::make_pair(domain_fqdn, ScanTaskInfo{domain_uuid, scan_type, {nameserver}}));
        }
    }
    return scan_tasks;
}

ScanTasks get_scan_tasks(const KeysetMetadata& _keyset_metadata, ExportScanType _export_scan_type)
{
    if (_keyset_metadata.zones.empty())
    {
        return ScanTasks{};
    }
    LibFred::OperationContextCreator ctx;
    const auto sql = construct_export_query(_keyset_metadata, _export_scan_type);
    const auto dbres = ctx.get_conn().exec_params(sql);
    ctx.commit_transaction();
    return get_scan_tasks_from_query_result(dbres);
}

} // namespace Fred::Akm::{anonymous}

void export_scan_tasks(
        ExportScanType _export_scan_type,
        const KeysetMetadata& _keyset_metadata,
        const SupervisorConnectionString& _supervisor_connection_string)
{
    try
    {
        const auto scan_tasks = get_scan_tasks(_keyset_metadata, _export_scan_type);
        GrpcClient::export_scan_tasks(_supervisor_connection_string, scan_tasks);
    }
    catch (const ::Database::Exception& e)
    {
        LIBLOG_ERROR("database exception: {}", e.what());
        throw;
    }
    catch (const GrpcException& e)
    {
        LIBLOG_ERROR("cdnskey-processor client exception: {}", e.what());
        throw;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error: {}", e.what());
        throw;
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown error");
        throw;
    }
}

} // namespace Fred::Akm
} // namespace Fred
