/*
 * Copyright (C) 2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/grpc_client/proto_unwrap.hh"

#include <algorithm>

namespace Fred {
namespace Akm {

ProtoFieldMissingException::ProtoFieldMissingException(const std::string& _proto, const std::string& _field)
    : ProtoException("in proto " + _proto + " field " + _field + " is missing")
{ }

namespace Api = CdnskeyProcessor::Api;

template <>
struct ProtoTypeTraits<Api::Supervisor::AcceptableDnskey>
{
    using InternalType = AcceptableKeyset;
};

template <>
struct ProtoTypeTraits<Api::Supervisor::Dnskey>
{
    using InternalType = Dnskey;
};

template <>
struct ProtoTypeTraits<Api::DomainFqdn>
{
    using InternalType = DomainFqdn;
};

template <>
struct ProtoTypeTraits<Api::DomainUid>
{
    using InternalType = DomainUuid;
};

template <>
DomainFqdn proto_unwrap(const Api::DomainFqdn& _src)
{
    return DomainFqdn{_src.value()};
}

template <>
DomainUuid proto_unwrap(const Api::DomainUid& _src)
{
    try
    {
        return from_string<DomainUuid>(_src.value());
    }
    catch (const std::runtime_error& e)
    {
        throw ProtoException(e.what());
    }
}

template <>
Dnskey proto_unwrap(const Api::Supervisor::Dnskey& _src)
{
    struct DnskeyFieldMissing : ProtoFieldMissingException
    {
        DnskeyFieldMissing(const std::string& _field)
            : ProtoFieldMissingException("Dnskey", _field)
        {
        }
    };
    if (!_src.has_flags())
    {
        throw DnskeyFieldMissing("flags");
    }
    if (!_src.has_proto())
    {
        throw DnskeyFieldMissing("proto");
    }
    if (!_src.has_alg())
    {
        throw DnskeyFieldMissing("alg");
    }
    if (!_src.has_public_key())
    {
        throw DnskeyFieldMissing("public_key");
    }
    return Dnskey{
        DnskeyFlags{_src.flags().value()},
        DnskeyProto{_src.proto().value()},
        DnskeyAlg{_src.alg().value()},
        DnskeyPublicKey{_src.public_key().value()}
    };
}

template <>
AcceptableKeyset proto_unwrap(const Api::Supervisor::AcceptableDnskey& _src)
{
    return AcceptableKeyset{
        proto_unwrap(_src.domain_uid()),
        {
            proto_unwrap(_src.domain_fqdn()),
            { proto_unwrap(_src.dnskey()) }
        }
    };
}

template <>
AcceptableKeysets proto_unwrap(const Api::Supervisor::AcceptableDnskeysReply& _src)
{
    AcceptableKeysets dst;
    std::for_each(_src.data().items().cbegin(), _src.data().items().cend(), [&dst](auto&& item)
    {
        append(dst, proto_unwrap(item));
    });
    return dst;
}

} // namespace Fred::Akm
} // namespace Fred
